/*****************************************************************************
 * Copyright (C) 2001, 2002 VideoLAN
 *
 * Authors: Christophe Massiot <massiot@via.ecp.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

// 08-Aug-2003: Edited by bluecow ( http://www.emule-project.net ); 
// 				based on ./vlc-0.6.1/modules/access/file.c from VideoLAN v0.6.1;
//				for detailed changelog see CHANGELOG.htm file

//#define USE_VLC_VER 0x062	// VLC 0.6.2
//#define USE_VLC_VER 0x070	// VLC 0.7.0
//#define USE_VLC_VER 0x071	// VLC 0.7.1
#define USE_VLC_VER 0x072	// VLC 0.7.2

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <vlc/vlc.h>
#include <vlc/input.h>

#include <stdlib.h>
#include <string.h>
#ifdef HAVE_SYS_TYPES_H
#   include <sys/types.h>
#endif
#ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#endif
#ifdef HAVE_SYS_STAT_H
#   include <sys/stat.h>
#endif
#ifdef HAVE_ERRNO_H
#   include <errno.h>
#endif
#ifdef HAVE_FCNTL_H
#   include <fcntl.h>
#endif

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#elif defined( WIN32 ) && !defined( UNDER_CE )
#   include <io.h>
#endif

/* stat() support for large files on win32 */
#if defined( WIN32 ) && !defined( UNDER_CE )
#define stat _stati64
#define fstat(a,b) _fstati64(a,b)
#endif

#ifdef UNDER_CE
#   define close(a) CloseHandle((HANDLE)a)
#endif

#if defined( WIN32 )
#include <share.h>
#endif

#ifdef _DEBUG
#define DEBUG_ONLY(f)      (f)
#else
#define DEBUG_ONLY(f)      ((void)0)
#endif

/*****************************************************************************
 * Exported prototypes
 *****************************************************************************/
static int     Open   ( vlc_object_t * );
static void    Close  ( vlc_object_t * );

static void    Seek   ( input_thread_t *, off_t );
static ssize_t Read   ( input_thread_t *, byte_t *, size_t );

static int     _OpenFile( input_thread_t *, char * );

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
#define CACHING_TEXT N_("Caching value in ms")
#define CACHING_LONGTEXT N_( \
    "Allows you to modify the default caching value for file streams. This " \
    "value should be set in millisecond units." )
#define CAT_TEXT N_("Concatenate with additional files")
#define CAT_LONGTEXT N_( \
    "Allows you to play split files as if they were part of a unique file. " \
    "Specify a comma-separated list of files." )

vlc_module_begin();
    set_description( _("eMule part file input") );
#if USE_VLC_VER<0x070
    add_category_hint( N_("file"), NULL, VLC_TRUE );
#endif
	add_usage_hint( N_("Use for previewing eMule part files") );

	// ------------
	add_bool( "mpg_partfile_stream", VLC_TRUE, NULL, 
			  N_("Enable plugin for MPEG part files"), 
			  N_("Specifies if this plugin shall provide a media stream with the information found in "
			     "an eMule part.met file."), 
			  VLC_FALSE );
	
	add_bool( "mpg_partfile_seekable", VLC_TRUE, NULL, 
			  N_("Provide a seekable stream for MPEG part files"), 
			  N_("Specifies if the MPEG media stream built from the eMule part file will be seekable."), 
			  VLC_FALSE );
	
	// ------------
	add_bool( "avi_partfile_stream", VLC_TRUE, NULL, 
			  N_("Enable plugin for AVI part files"), 
			  N_("Specifies if this plugin shall provide a media stream with the information found in "
			     "an eMule part.met file."), 
			  VLC_FALSE );

	add_bool( "avi_partfile_seekable", VLC_FALSE, NULL, 
			  N_("Provide a seekable stream for AVI part files"), 
			  N_("Specifies if the AVI media stream built from the eMule part file will be seekable."), 
			  VLC_FALSE );
	
	// ------------
	add_bool( "oth_partfile_stream", VLC_TRUE, NULL, 
			  N_("Enable plugin for other (non MPEG or AVI) part files"), 
			  N_("Specifies if this plugin shall provide a media stream with the information found in "
			     "an eMule part.met file."), 
			  VLC_FALSE );

	add_bool( "oth_partfile_seekable", VLC_FALSE, NULL, 
			  N_("Provide a seekable stream for other (non MPEG or AVI) part files"), 
			  N_("Specifies if the media stream built from the eMule part file will be seekable."), 
			  VLC_FALSE );
	
	// ------------
	add_bool( "handle_allfiles", VLC_FALSE, NULL, 
			  N_("Enable plugin for all type of files (eMule part files and regular files)"), 
			  N_("Specifies if this plugin shall handle all type of files regardless whether it is "
			     "an eMule part file or a regular file. If this option is disabled, this plugin will "
				 "handle only eMule part files."), 
			  VLC_FALSE );

    add_integer( "file-caching", DEFAULT_PTS_DELAY / 1000, NULL, CACHING_TEXT, CACHING_LONGTEXT, VLC_TRUE );
    add_string( "file-cat", NULL, NULL, CAT_TEXT, CAT_LONGTEXT, VLC_TRUE );

    set_capability( "access", 51 );	// >50 to be called before the 'access_file' module
    add_shortcut( "partfile" );		// can be used for an access protocol e.g. "partfile://D:\Temp\001.part"
    add_shortcut( "file" );			// can be used for an access protocol e.g. "file://D:\Temp\001.part"
    set_callbacks( Open, Close );

// 07-Jn-2004 [bc]: The 'vlc_module_end' macro of VLC 0.7.0 can not be compiled with VC7
#if USE_VLC_VER!=0x070 // use for VLC 0.6.2, VLC 0.7.1(+)

vlc_module_end();

#else

            p_submodule->pp_shortcuts[ i_shortcut ] = NULL;
        }
        if( p_config )
        {
			module_config_t __module_config___CONFIG_HINT_END = { CONFIG_HINT_END };
            p_config[ ++i_config ] = __module_config___CONFIG_HINT_END;
            config_Duplicate( p_module, p_config );
            free( p_config );
        }
        else
		{
			module_config_t __module_config___CONFIG_HINT_END = { CONFIG_HINT_END };
			config_Duplicate(p_module, &__module_config___CONFIG_HINT_END);
		}
        if( p_module->p_config == NULL )
        {
            return VLC_EGENERIC;
        }
        return VLC_SUCCESS && i_shortcut;
    }
    struct _u_n_u_s_e_d_ /* the ; gets added */
	;

#endif

/*****************************************************************************
 * _input_socket_t: private access plug-in data, modified to add private
 *                  fields
 *****************************************************************************/

typedef struct
{
	uint32_t		l_start;
	uint32_t		l_end;
} GAP;

typedef struct
{
	uint32_t		l_start;
	uint32_t		l_end;
	uint32_t		l_stream_start;
	uint32_t		l_stream_end;
	int				i_next_gap;
} AVAIL;

enum EFileType
{
	ftNUL = 0,
	ftMPG,
	ftAVI
};

typedef struct file_entry_s
{
    char     *psz_name;
    int64_t  i_size;

} file_entry_t;

typedef struct _input_socket_s
{
    input_socket_t      _socket;

    unsigned int        i_nb_reads;
    vlc_bool_t          b_kfir;

    /* Files list */
    file_entry_t **p_files;
    int          i_files;

    /* Current file */
    int  i_index;

	int					i_read_errors;

	uint32_t			l_part_file_size;
	uint32_t			l_avail_size;

	int					i_gaps;
	GAP *				p_gaps;

	int					i_avails;
	AVAIL *				p_avails;

	int					i_next_gap;
	unsigned long		l_fpos;		//TODO: maybe we could use "input_area_t.i_tell" instead of "l_fpos"

	enum EFileType		n_file_type;
} _input_socket_t;

#define PARTFILE_VERSION		0xe0
//file tags
#define FT_FILENAME				0x01	// <string>
#define FT_FILESIZE				0x02	// <int>
#define FT_FILETYPE				0x03	// <string>
#define FT_FILEFORMAT			0x04	// <string>
#define FT_GAPSTART				0x09
#define FT_GAPEND				0x0A
#define	FT_MEDIA_ARTIST			0xD0	// <string>
#define	FT_MEDIA_ALBUM			0xD1	// <string>
#define	FT_MEDIA_TITLE			0xD2	// <string>
#define	FT_MEDIA_LENGTH			0xD3	// <uint32> !!!
#define	FT_MEDIA_BITRATE		0xD4	// <uint32>
#define	FT_MEDIA_CODEC			0xD5	// <string>

static int part_met_file_error(input_thread_t* p_input, const char* psz_format, ...)
{
	char sz_msg[1024];
	va_list argp;
	va_start(argp, psz_format);
	vsnprintf(sz_msg, sizeof sz_msg, psz_format, argp);
	va_end(argp);
	msg_Err(p_input, "%s", sz_msg);
	return 0;
}

static int CmpGaps(const void* p0, const void* p1)
{
	if( ((const GAP*)p0)->l_start < ((const GAP*)p1)->l_start )
		return -1;
	if( ((const GAP*)p0)->l_start > ((const GAP*)p1)->l_start )
		return 1;
	return 0;
}

// VLC 0.6.2 does not display strings which contain characters >=0x80 ( because it expects UTF encoded chars )
// VLC 0.7.0 does handle Windows ANSI codepage characters as expected
#if USE_VLC_VER<0x070
static void str2ascii(char* psz)
{
	while( *psz != '\0' )
	{
		if( !isascii(*psz) )
		{
			     if( *psz == '' )	*psz = 'a';
			else if( *psz == '' )	*psz = 'o';
			else if( *psz == '' )	*psz = 'u';
			else if( *psz == '' )	*psz = 'A';
			else if( *psz == '' )	*psz = 'O';
			else if( *psz == '' )	*psz = 'U';
			else                    *psz = '.';
		}
		psz++;
	}
}
#endif

static char* strdup2utf(char* psz)
{
	char* psza;
#if USE_VLC_VER<0x070
	size_t n_maxwsz = MultiByteToWideChar(CP_ACP, 0, psz, -1, NULL, 0);
	if( n_maxwsz != 0 )
	{
		wchar_t* pwsz = (wchar_t*)malloc(sizeof(*pwsz) * n_maxwsz);
		if( pwsz != NULL )
		{
			size_t n_cnv = MultiByteToWideChar(CP_ACP, 0, psz, -1, pwsz, n_maxwsz);
			if( n_cnv == n_maxwsz )
			{
				size_t n_maxutf = WideCharToMultiByte(CP_UTF8, 0, pwsz, -1, NULL, 0, NULL, NULL);
				char* putf = (char*)malloc(sizeof(*putf) * n_maxutf);
				if( putf != NULL )
				{
					n_cnv = WideCharToMultiByte(CP_UTF8, 0, pwsz, -1, putf, n_maxutf, NULL, NULL);
					if( n_cnv == n_maxutf )
					{
						free(pwsz);
						return putf;
					}
					free(putf);
				}
			}
			free(pwsz);
		}
	}
	// If all failed, do at least some kind of ASCII cleanup
#endif
	psza = _strdup(psz);
#if USE_VLC_VER<0x070
	str2ascii(psza);
#endif
	return psza;
}

static void SecToTimeLength(uint32_t l_sec, char* buff, size_t size)
{
	if( l_sec >= 3600 )
	{
		unsigned uHours = l_sec/3600;
		unsigned uMin = (l_sec - uHours*3600)/60;
		unsigned uSec = l_sec - uHours*3600 - uMin*60;
		_snprintf(buff, size, "%u:%02u:%02u", uHours, uMin, uSec);
	}
	else
	{
		unsigned uMin = l_sec/60;
		unsigned uSec = l_sec - uMin*60;
		_snprintf(buff, size, "%u:%02u", uMin, uSec);
	}
}

static int read_part_met_file(input_thread_t* p_input, _input_socket_t* a, FILE* file)
{
	uint8_t c_version;
	uint16_t n_part_hashs;
	uint32_t l_meta_tags;
	uint32_t l_tag;
	uint32_t l_gap_start;
	uint32_t l_total_missing;
	uint32_t l_last_end;
	uint32_t l_stream_pos;
	int i;
	char sz_filename[_MAX_PATH];
	char sz_fileformat[_MAX_PATH];
	const char* psz_format = NULL;
	uint32_t n_media_length;
	uint32_t n_media_bitrate;
	char* psz_media_artist = NULL;
	char* psz_media_album = NULL;
	char* psz_media_title = NULL;

	// BYTE: Version
	if( fread(&c_version, sizeof(c_version), 1, file) != 1 )
		return part_met_file_error(p_input, "failed to read file version");
	if( c_version != PARTFILE_VERSION )
		return part_met_file_error(p_input, "file version != 0x%02x", PARTFILE_VERSION);

	// INT32: Date
	if( fseek(file, 4, SEEK_CUR) != 0 )
		return part_met_file_error(p_input, "failed to read date");

	// BYTE*16: File ID
	if( fseek(file, 16, SEEK_CUR) != 0 )
		return part_met_file_error(p_input, "failed to read file hash");

	// UINT16: Nr. of part hashs
	if( fread(&n_part_hashs, sizeof(n_part_hashs), 1, file) != 1 )
		return part_met_file_error(p_input, "failed to read number of part hashs");

	// <nr> * (BYTE*16): part hashs
	if( fseek(file, (unsigned)n_part_hashs*16, SEEK_CUR) != 0 )
		return part_met_file_error(p_input, "failed to read part hashs");

	// INT32: nr. meta tags
	if( fread(&l_meta_tags, sizeof(l_meta_tags), 1, file) != 1 )
		return part_met_file_error(p_input, "failed to read number of meta tags");

	l_gap_start = 0;
	l_total_missing = 0;
	sz_filename[0] = '\0';
	sz_fileformat[0] = '\0';
	n_media_length = 0;
	n_media_bitrate = 0;
	for( l_tag = 0; l_tag < l_meta_tags; l_tag++ )
	{
		uint8_t c_tag_type;
		uint16_t n_name_length;
		char sz_name[4096];
		long l_file_pos = ftell(file);

		if( fread(&c_tag_type, sizeof(c_tag_type), 1, file) != 1 )
		{
			part_met_file_error(p_input, "failed to read type of tag #%u", l_tag+1);
			return 0;
		}

		if( fread(&n_name_length, sizeof(n_name_length), 1, file) != 1 )
			return part_met_file_error(p_input, "failed to read name length of tag #%u", l_tag+1);
		if( fread(sz_name, 1, n_name_length, file) != (size_t)n_name_length )
			return part_met_file_error(p_input, "failed to read tag name of tag #%u", l_tag+1);
		sz_name[n_name_length] = '\0';

		if (c_tag_type == 2) // String value
		{
			uint16_t n_value_length;
			char sz_value[4096];
			
			if( fread(&n_value_length, sizeof(n_value_length), 1, file) != 1 )
				return part_met_file_error(p_input, "failed to read string value length of tag #%u", l_tag+1);
			if( fread(sz_value, 1, n_value_length, file) != (size_t)n_value_length )
				return part_met_file_error(p_input, "failed to read string value of tag #%u", l_tag+1);
			sz_value[n_value_length] = '\0';

			if( n_name_length == 1 )
			{
				switch( (unsigned char)sz_name[0] )
				{
				case FT_FILENAME:
					strncpy(sz_filename, sz_value, sizeof(sz_filename));
					sz_filename[sizeof(sz_filename) - 1] = '\0';
					break;
				case FT_FILEFORMAT:
					strncpy(sz_fileformat, sz_value, sizeof(sz_fileformat));
					sz_fileformat[sizeof(sz_fileformat) - 1] = '\0';
					break;
				case FT_MEDIA_ARTIST:
					psz_media_artist = _strdup(sz_value);
					break;
				case FT_MEDIA_ALBUM:
					psz_media_album = _strdup(sz_value);
					break;
				case FT_MEDIA_TITLE:
					psz_media_title = _strdup(sz_value);
					break;
				default:
					break;
				}
			}
		}
		else if( c_tag_type == 3 ) // Integer value
		{
			uint32_t n_value;

			if( fread(&n_value, sizeof(n_value), 1, file) != 1 )
				return part_met_file_error(p_input, "failed to read integer value of tag #%u", l_tag+1);

			if( n_name_length == 1 )
			{
				switch( (unsigned char)sz_name[0] )
				{
				case FT_FILESIZE:
					a->l_part_file_size = n_value;
					break;
				case FT_MEDIA_LENGTH:
					n_media_length = n_value;
					break;
				case FT_MEDIA_BITRATE:
					n_media_bitrate = n_value;
					break;
				}
			}
			else
			{
				if( sz_name[0] == FT_GAPSTART )
				{
					l_gap_start = n_value;
					a->p_gaps = (GAP*)realloc(a->p_gaps, sizeof(*a->p_gaps) * (a->i_gaps + 1));
					a->p_gaps[a->i_gaps].l_start = n_value;
					a->p_gaps[a->i_gaps].l_end = -1; // just in case
				}
				else if( sz_name[0] == FT_GAPEND )
				{
					l_total_missing += n_value - l_gap_start;
					a->p_gaps[a->i_gaps].l_end = n_value;
					a->i_gaps++;
				}
			}
		}
		else if( c_tag_type == 4 ) // Float value
		{
			float f_value;

			if( fread(&f_value, sizeof(f_value), 1, file) != 1 )
				return part_met_file_error(p_input, "failed to read float value of tag #%u", l_tag+1);
		}
		else
		{
            part_met_file_error(p_input, "unknown tag type %u at 0x%08x", c_tag_type, l_file_pos);
			return 0;
		}
	}

	a->l_avail_size = a->l_part_file_size - l_total_missing;

	qsort(a->p_gaps, a->i_gaps, sizeof(*a->p_gaps), CmpGaps);
	
#ifdef _DEBUG
	for( i = 0; i < a->i_gaps; i++ )
		msg_Info( p_input, "gap %3u: %10u-%10u = %10u (%7.2f MByte)", i, 
			a->p_gaps[i].l_start, a->p_gaps[i].l_end, 
			a->p_gaps[i].l_end - a->p_gaps[i].l_start, (double)(a->p_gaps[i].l_end - a->p_gaps[i].l_start) / (1024.0 * 1024.0));
#endif

	l_stream_pos = 0;
	l_last_end = 0;
	for( i = 0; i < a->i_gaps; i++ )
	{
		if( a->p_gaps[i].l_start > l_last_end )
		{
			a->p_avails = (AVAIL*)realloc(a->p_avails, sizeof(*a->p_avails) * (a->i_avails + 1));
			a->p_avails[a->i_avails].l_start = l_last_end;
			a->p_avails[a->i_avails].l_end = a->p_gaps[i].l_start;
			a->p_avails[a->i_avails].l_stream_start = l_stream_pos;
			a->p_avails[a->i_avails].l_stream_end = a->p_avails[a->i_avails].l_stream_start + (a->p_avails[a->i_avails].l_end - a->p_avails[a->i_avails].l_start);
			a->p_avails[a->i_avails].i_next_gap = a->i_gaps; // just in case
			l_stream_pos += a->p_avails[a->i_avails].l_end - a->p_avails[a->i_avails].l_start;
			a->i_avails++;
		}
		l_last_end = a->p_gaps[i].l_end;
	}
	if( a->l_part_file_size > l_last_end )
	{
		a->p_avails = (AVAIL*)realloc(a->p_avails, sizeof(*a->p_avails) * (a->i_avails + 1));
		a->p_avails[a->i_avails].l_start = l_last_end;
		a->p_avails[a->i_avails].l_end = a->l_part_file_size;
		a->p_avails[a->i_avails].l_stream_start = l_stream_pos;
		a->p_avails[a->i_avails].l_stream_end = a->p_avails[a->i_avails].l_stream_start + (a->p_avails[a->i_avails].l_end - a->p_avails[a->i_avails].l_start);
		a->i_avails++;
	}

#ifdef _DEBUG
	for( i = 0; i < a->i_avails; i++ )
		msg_Info( p_input, "avl %3u: %10u-%10u  %10u-%10u = %10u (%7.2f MByte)", i, 
			a->p_avails[i].l_start, a->p_avails[i].l_end, 
			a->p_avails[i].l_stream_start, a->p_avails[i].l_stream_end,
			a->p_avails[i].l_end - a->p_avails[i].l_start, (double)(a->p_avails[i].l_end - a->p_avails[i].l_start) / (1024.0 * 1024.0));
#endif

	// for each 'avail' section pre-determine the 'next_gap'
	for( i = 0; i < a->i_avails; i++ )
	{
		int j;

		a->p_avails[i].i_next_gap = a->i_gaps;
		for(j = 0; j < a->i_gaps; j++ )
		{
			if( a->p_avails[i].l_start <= a->p_gaps[j].l_start )
			{
				a->p_avails[i].i_next_gap = j;
				break;
			}
		}
	}

	a->n_file_type = ftNUL;
	psz_format = NULL;
	// give the filename a higher priority when determining the file type.. although the information
	// stored in FT_FILEFORMAT should be more accurat, the filename is the only way the user can 
	// change/specify the file type!
	if( sz_filename[0] != '\0' )
		psz_format = strrchr(sz_filename, '.');
	if( !psz_format && sz_fileformat[0] != '\0' )
		psz_format = sz_fileformat;

	if( psz_format )
	{
		// trim leading spaces and dots
		while( *psz_format == ' ' || *psz_format == '.' )
			psz_format++;

		if( *psz_format != '\0' )
		{
			char* psz;
			char sz_format[16];

			strncpy(sz_format, psz_format, sizeof(sz_format));
			sz_format[sizeof(sz_format) - 1] = '\0';

			// trim trailing spaces and dots
			psz = sz_format + 1;
			while( *psz != '\0' )
			{
				if( *psz == ' ' || *psz == '.' )
				{
					*psz = '\0';
					break;
				}
				psz++;
			}

			_strlwr(sz_format);
			if(    !strcmp(sz_format, "mpg")
				|| !strcmp(sz_format, "mpeg")
				|| !strcmp(sz_format, "mpe")
				|| !strcmp(sz_format, "mpv")
				|| !strcmp(sz_format, "mpv1")
				|| !strcmp(sz_format, "mpv2")
				|| !strcmp(sz_format, "mpv3")
				|| !strcmp(sz_format, "mpv4")
				|| !strcmp(sz_format, "mp1v")
				|| !strcmp(sz_format, "mp2v")
				|| !strcmp(sz_format, "mp3v")
				|| !strcmp(sz_format, "mp4v")
				|| !strcmp(sz_format, "mpa")
				|| !strcmp(sz_format, "mp1")
				|| !strcmp(sz_format, "mp2")
				|| !strcmp(sz_format, "mp3")
				|| !strcmp(sz_format, "mp4")
				|| !strcmp(sz_format, "mps") )
				a->n_file_type = ftMPG;
			else if( !strcmp(sz_format, "avi") )
				a->n_file_type = ftAVI;
		}
	}

#if USE_VLC_VER<=0x071
	if( p_input->stream.p_info )
	{
		input_info_category_t* p_info_cat = p_input->stream.p_info;

		while( p_info_cat )
		{
			if( !p_info_cat->p_next )
			{
				input_info_category_t* p_met_info_cat = calloc(1, sizeof(*p_met_info_cat));

				if( p_met_info_cat )
				{
					input_info_t ** pp_next;

					p_met_info_cat->psz_name = strdup2utf("eMule part file meta data");
					p_info_cat->p_next = p_met_info_cat;
					pp_next = &p_met_info_cat->p_info;

					if( sz_filename[0] != '\0' )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							p_met_info->psz_name = strdup2utf("Name");
							p_met_info->psz_value = strdup2utf(sz_filename);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( a->l_part_file_size > 0 )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							char sz_data[64];
							_snprintf(sz_data, sizeof sz_data, "%u bytes (%.2f Mbyte)", a->l_part_file_size, a->l_part_file_size/(1024.0*1024.0));
							p_met_info->psz_name = strdup2utf("Size");
							p_met_info->psz_value = strdup2utf(sz_data);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( a->l_avail_size > 0 )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							char sz_data[64];
							_snprintf(sz_data, sizeof sz_data, "%u bytes (%.2f Mbyte)", a->l_avail_size, a->l_avail_size/(1024.0*1024.0));
							p_met_info->psz_name = strdup2utf("Available data");
							p_met_info->psz_value = strdup2utf(sz_data);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( sz_fileformat[0] != '\0' )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							p_met_info->psz_name = strdup2utf("Format");
							p_met_info->psz_value = strdup2utf(sz_fileformat);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( n_media_bitrate > 0 )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							char sz_data[32];
							_snprintf(sz_data, sizeof sz_data, "%u kBit/s", n_media_bitrate);
							p_met_info->psz_name = strdup2utf("Bitrate");
							p_met_info->psz_value = strdup2utf(sz_data);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( n_media_length > 0 )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							char sz_data[32];
							SecToTimeLength(n_media_length, sz_data, sizeof sz_data);
							p_met_info->psz_name = strdup2utf("Length");
							p_met_info->psz_value = strdup2utf(sz_data);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( psz_media_artist )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							p_met_info->psz_name = strdup2utf("Artist");
							p_met_info->psz_value = strdup2utf(psz_media_artist);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( psz_media_album )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							p_met_info->psz_name = strdup2utf("Album");
							p_met_info->psz_value = strdup2utf(psz_media_album);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}

					if( psz_media_title )
					{
						input_info_t * p_met_info = calloc(1, sizeof(*p_met_info));
						if( p_met_info )
						{
							p_met_info->psz_name = strdup2utf("Title");
							p_met_info->psz_value = strdup2utf(psz_media_title);
							*pp_next = p_met_info;
							pp_next = &p_met_info->p_next;
						}
					}
				}

				break;
			}
			p_info_cat = p_info_cat->p_next;
		}
	}
#else
	/* Can not be used for VLC 0.7.2 (see comments at end of 'if' scope) */
	if( p_input->p_item )
	{
		info_category_t* p_met_info_cat = calloc(1, sizeof(*p_met_info_cat));

		if( p_met_info_cat )
		{
			p_met_info_cat->psz_name = strdup2utf("eMule part file meta data");

			if( sz_filename[0] != '\0' )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					p_met_info->psz_name = strdup2utf("Name");
					p_met_info->psz_value = strdup2utf(sz_filename);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( a->l_part_file_size > 0 )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					char sz_data[64];
					_snprintf(sz_data, sizeof sz_data, "%u bytes (%.2f Mbyte)", a->l_part_file_size, a->l_part_file_size/(1024.0*1024.0));
					p_met_info->psz_name = strdup2utf("Size");
					p_met_info->psz_value = strdup2utf(sz_data);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( a->l_avail_size > 0 )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					char sz_data[64];
					_snprintf(sz_data, sizeof sz_data, "%u bytes (%.2f Mbyte)", a->l_avail_size, a->l_avail_size/(1024.0*1024.0));
					p_met_info->psz_name = strdup2utf("Available data");
					p_met_info->psz_value = strdup2utf(sz_data);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( sz_fileformat[0] != '\0' )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					p_met_info->psz_name = strdup2utf("Format");
					p_met_info->psz_value = strdup2utf(sz_fileformat);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( n_media_bitrate > 0 )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					char sz_data[32];
					_snprintf(sz_data, sizeof sz_data, "%u kBit/s", n_media_bitrate);
					p_met_info->psz_name = strdup2utf("Bitrate");
					p_met_info->psz_value = strdup2utf(sz_data);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( n_media_length > 0 )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					char sz_data[32];
					SecToTimeLength(n_media_length, sz_data, sizeof sz_data);
					p_met_info->psz_name = strdup2utf("Length");
					p_met_info->psz_value = strdup2utf(sz_data);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( psz_media_artist )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					p_met_info->psz_name = strdup2utf("Artist");
					p_met_info->psz_value = strdup2utf(psz_media_artist);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( psz_media_album )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					p_met_info->psz_name = strdup2utf("Album");
					p_met_info->psz_value = strdup2utf(psz_media_album);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			if( psz_media_title )
			{
				info_t * p_met_info = calloc(1, sizeof(*p_met_info));
				if( p_met_info )
				{
					p_met_info->psz_name = strdup2utf("Title");
					p_met_info->psz_value = strdup2utf(psz_media_title);
					p_met_info_cat->pp_infos = realloc(p_met_info_cat->pp_infos, sizeof(*p_met_info_cat->pp_infos) * (p_met_info_cat->i_infos + 1));
					if( p_met_info_cat->pp_infos )
						p_met_info_cat->pp_infos[p_met_info_cat->i_infos++] = p_met_info;
				}
			}

			// Problem: We can not use the MSVCRT heap memory functions for reallocating (or even freeing) the ptrs
			// which were already allocated by VLC core (would be needed to append a category). MSVCRT complains
			// about that those ptrs are using a different heap! So, all that category code would work if we
			// would have the same heap as VLC core, but for any (to me unknown) reason we don't have it.
			p_input->p_item->pp_categories = realloc(p_input->p_item->pp_categories, sizeof(*p_input->p_item->pp_categories) * (p_input->p_item->i_categories + 1));
			if( p_input->p_item->pp_categories )
				p_input->p_item->pp_categories[p_input->p_item->i_categories++] = p_met_info_cat;
		}
	}
#endif
	if( psz_media_artist )
		free(psz_media_artist);
	if( psz_media_album )
		free(psz_media_album);
	if( psz_media_title )
		free(psz_media_title);

	return 1;
}

/*****************************************************************************
 * Open: open the file
 *****************************************************************************/
static int Open( vlc_object_t *p_this )
{
    input_thread_t *    p_input = (input_thread_t *)p_this;
    char *              psz_name = p_input->psz_name;
#ifdef HAVE_SYS_STAT_H
    int                 i_stat;
    struct stat         stat_info;
#endif
    _input_socket_t *   p_access_data;
    vlc_bool_t          b_stdin, b_kfir = 0;
    vlc_value_t         val;
    file_entry_t *      p_file;
	char				sz_partmet_file[_MAX_PATH];
	char *				psz_ext = NULL;
	FILE *				fp = NULL;

    p_input->i_mtu = 0;

    b_stdin = psz_name[0] == '-' && psz_name[1] == '\0';

#ifdef HAVE_SYS_STAT_H
    if( !b_stdin && (i_stat = stat( psz_name, &stat_info )) == (-1) )
    {
#   ifdef HAVE_ERRNO_H
        msg_Warn( p_input, "cannot stat() file `%s' (%s)",
                  psz_name, strerror(errno));
#   else
        msg_Warn( p_input, "cannot stat() file `%s'", psz_name );
#   endif
        return VLC_EGENERIC;
    }
#endif

    p_access_data = malloc( sizeof(_input_socket_t) );
    if( p_access_data == NULL )
    {
        msg_Err( p_input, "out of memory" );
        return VLC_ENOMEM;
    }
	memset(p_access_data, 0, sizeof(*p_access_data));

	// the idea is:
	// 1.) handle part files with part.met files
	// 2.) even if the reading of the part.met file failed, handle the file to get around some nasty bug in VLC
	//	   which freezes the entire box when reading a file in an endless loop :(

	// if it's a part file...
	if(    config_GetInt(p_input, "mpg_partfile_stream")
		|| config_GetInt(p_input, "avi_partfile_stream")
		|| config_GetInt(p_input, "oth_partfile_stream") )
	{
		psz_ext = strrchr(psz_name, '.');
		if( psz_ext != NULL && strcasecmp(psz_ext, ".part") == 0 )
		{
			strcpy(sz_partmet_file, psz_name);
			strcat(sz_partmet_file, ".met");

			// ... try to read the part.met file
			fp = _fsopen(sz_partmet_file, "rb", _SH_DENYWR);
			if( fp != NULL )
			{
				__try
				{
					if(    !read_part_met_file(p_input, p_access_data, fp)
						|| (p_access_data->n_file_type == ftAVI && !config_GetInt(p_input, "avi_partfile_stream"))
						|| (p_access_data->n_file_type == ftMPG && !config_GetInt(p_input, "mpg_partfile_stream"))
						|| (p_access_data->n_file_type == ftNUL && !config_GetInt(p_input, "oth_partfile_stream")) )
					{
						// if the part.met file reading failed, handle the part file as a normal file (see point 2 above)
						if( p_access_data->p_gaps )
						{
							free(p_access_data->p_gaps);
							p_access_data->p_gaps = NULL;
						}
						if( p_access_data->p_avails )
						{
							free(p_access_data->p_avails);
							p_access_data->p_avails = NULL;
						}
						p_access_data->i_gaps = 0;
						p_access_data->i_avails = 0;
						p_access_data->i_next_gap = 0;
						p_access_data->l_avail_size = 0;
						p_access_data->l_fpos = 0;
						p_access_data->l_part_file_size = 0;
					}
				}
				__except(EXCEPTION_EXECUTE_HANDLER)
				{
					char sz_msg[128 + MAX_PATH];
					_snprintf(sz_msg, sizeof sz_msg, "Unhandled exception while reading \"%s\"", sz_partmet_file);
					MessageBox(HWND_DESKTOP, sz_msg, "eMule part file plugin", MB_OK | MB_ICONSTOP | MB_SYSTEMMODAL | MB_SETFOREGROUND);
					// don't try to free anything after an exception
					p_access_data->p_gaps = NULL;
					p_access_data->p_avails = NULL;
					p_access_data->i_gaps = 0;
					p_access_data->i_avails = 0;
					p_access_data->i_next_gap = 0;
					p_access_data->l_avail_size = 0;
					p_access_data->l_fpos = 0;
					p_access_data->l_part_file_size = 0;
				}
				fclose(fp);
			}
			else
			{
				msg_Err( p_input, "failed to open part.met file `%s'", sz_partmet_file );
			}
		}
	}

	if( !p_access_data->p_gaps && !config_GetInt(p_input, "handle_allfiles") )
		return VLC_EGENERIC;
	msg_Info( p_input, "file is %shandled as part file", !p_access_data->p_gaps ? "not " : "");

    p_input->pf_read = Read;
    p_input->pf_set_program = input_SetProgram;
    p_input->pf_set_area = NULL;
	p_input->pf_seek = Seek;

    vlc_mutex_lock( &p_input->stream.stream_lock );

    if( *p_input->psz_access && !strncmp(p_input->psz_access, "stream", 7) )
    {
        /* stream:%s */
        p_input->stream.b_pace_control = 0;
        p_input->stream.b_seekable = 0;
        p_input->stream.p_selected_area->i_size = 0;
    }
	else
	{
		/* partfile:%s */
		/* file:%s or %s */
		p_input->stream.b_pace_control = 1;

        if( b_stdin )
        {
            p_input->stream.b_seekable = 0;
            p_input->stream.p_selected_area->i_size = 0;
        }
#ifdef UNDER_CE
        else if( VLC_TRUE )
        {
            /* We'll update i_size after it's been opened */
            p_input->stream.b_seekable = 1;
        }
#elif defined( HAVE_SYS_STAT_H )
		else if( S_ISREG(stat_info.st_mode) || S_ISCHR(stat_info.st_mode)
				  || S_ISBLK(stat_info.st_mode) )
		{
			if( p_access_data->p_gaps )
			{
				// offering the part file stream as a seekable stream does not always work.. sometimes there are
				// strange results.. sometimes there is no video shown at all, though the audio is played..

				if(    (p_access_data->n_file_type == ftAVI && config_GetInt(p_input, "avi_partfile_seekable"))
					|| (p_access_data->n_file_type == ftMPG && config_GetInt(p_input, "mpg_partfile_seekable"))
				    || (p_access_data->n_file_type == ftNUL && config_GetInt(p_input, "oth_partfile_seekable")) )
					p_input->stream.b_seekable = 1;
				else
					p_input->stream.b_seekable = 0;
			}
			else
				p_input->stream.b_seekable = 1;
			if( p_access_data->p_gaps )
				p_input->stream.p_selected_area->i_size = p_access_data->l_avail_size;
			else
				p_input->stream.p_selected_area->i_size = stat_info.st_size;
		}
        else if( S_ISFIFO(stat_info.st_mode)
#   if !defined( SYS_BEOS ) && !defined( WIN32 )
                  || S_ISSOCK(stat_info.st_mode)
#   endif
               )
        {
            p_input->stream.b_seekable = 0;
            p_input->stream.p_selected_area->i_size = 0;
        }
#endif
		else
		{
			vlc_mutex_unlock( &p_input->stream.stream_lock );
            msg_Err( p_input, "unknown file type for `%s'", psz_name );
			if( p_access_data->p_gaps )
				free(p_access_data->p_gaps);
			if( p_access_data->p_avails )
				free(p_access_data->p_avails);
			free(p_access_data);
			return VLC_EGENERIC;
		}
	}
 
    p_input->stream.p_selected_area->i_tell = 0;
    p_input->stream.i_method = INPUT_METHOD_FILE;
    vlc_mutex_unlock( &p_input->stream.stream_lock );
 
    msg_Dbg( p_input, "opening file `%s'", psz_name );
    p_input->p_access_data = (void *)p_access_data;

    p_access_data->i_nb_reads = 0;
    p_access_data->b_kfir = b_kfir;
    if( b_stdin )
    {
        p_access_data->_socket.i_handle = 0;
    }
    else
    {
        if( _OpenFile( p_input, psz_name ) != VLC_SUCCESS )
        {
		    if( p_access_data->p_gaps )
			    free( p_access_data->p_gaps );
		    if( p_access_data->p_avails )
			    free( p_access_data->p_avails );
            free(p_access_data);
            return VLC_EGENERIC;
        }
	}
	
	if ( (p_input->stream.b_seekable
          || p_access_data->p_gaps)
	 	 && !p_input->stream.p_selected_area->i_size )
    {
        msg_Err( p_input, "file %s is empty, aborting", psz_name );
		if( p_access_data->p_gaps )
			free( p_access_data->p_gaps );
		if( p_access_data->p_avails )
			free( p_access_data->p_avails );
        free(p_access_data);
        return VLC_EGENERIC;
    }

	if( p_access_data->p_gaps )
		msg_Info( p_input, "file contains %u bytes (%.2f Mbyte) of actual data; stream is %sseekable", p_access_data->l_avail_size, (float)p_access_data->l_avail_size/(1024*1024), p_input->stream.b_seekable ? "" : "not ");

	/* Update default_pts to a suitable value for file access */
#if USE_VLC_VER<0x070
    p_input->i_pts_delay = config_GetInt( p_input, "file-caching" ) * 1000;
#else
    var_Create( p_input, "file-caching", VLC_VAR_INTEGER | VLC_VAR_DOINHERIT );
    var_Get( p_input, "file-caching", &val );
    p_input->i_pts_delay = val.i_int * 1000;
#endif

    /*
     * Get the additional list of files
     */
    p_access_data->p_files = NULL;
    p_access_data->i_files = p_access_data->i_index = 0;
    p_file = (file_entry_t *)malloc( sizeof(file_entry_t) );
    p_file->i_size = p_input->stream.p_selected_area->i_size;
    p_file->psz_name = strdup( psz_name );
    TAB_APPEND( p_access_data->i_files, p_access_data->p_files, p_file );

    var_Create( p_input, "file-cat", VLC_VAR_STRING | VLC_VAR_DOINHERIT );
    var_Get( p_input, "file-cat", &val );

    if( val.psz_string && *val.psz_string )
    {
        char *psz_parser = psz_name = val.psz_string;
        int64_t i_size;

        while( psz_name && *psz_name )
        {
            psz_parser = strchr( psz_name, ',' );
            if( psz_parser ) *psz_parser = 0;

            psz_name = strdup( psz_name );
            if( psz_name )
            {
                msg_Dbg( p_input, "adding file `%s'", psz_name );

#ifdef HAVE_SYS_STAT_H
                if( !stat( psz_name, &stat_info ) )
                {
                    p_input->stream.p_selected_area->i_size +=
                        stat_info.st_size;
                    i_size = stat_info.st_size;
                }
                else
                {
                    msg_Dbg( p_input, "cannot stat() file `%s'", psz_name );
                    i_size = 0;
                }
#endif

                p_file = (file_entry_t *)malloc( sizeof(file_entry_t) );
                p_file->i_size = i_size;
                p_file->psz_name = psz_name;

                TAB_APPEND( p_access_data->i_files, p_access_data->p_files,
                            p_file );
            }

            psz_name = psz_parser;
            if( psz_name ) psz_name++;
        }
    }

	// Problem: We can not use the MSVCRT heap memory functions for reallocating (or even freeing) the ptrs
	// which were already allocated by VLC core. MSVCRT complains about that those ptrs are using a different heap!
	if( val.psz_string ) free( val.psz_string );

    return VLC_SUCCESS;
}

/*****************************************************************************
 * Close: close the target
 *****************************************************************************/
static void Close( vlc_object_t * p_this )
{
    input_thread_t * p_input = (input_thread_t *)p_this;
    _input_socket_t *p_access_data = (_input_socket_t *)p_input->p_access_data;
    int i;

    msg_Info( p_input, "closing `%s/%s://%s'", 
              p_input->psz_access, p_input->psz_demux, p_input->psz_name );
 
    close( p_access_data->_socket.i_handle );

    for( i = 0; i < p_access_data->i_files; i++ )
    {
        free( p_access_data->p_files[i]->psz_name );
        free( p_access_data->p_files[i] );
    }
    if( p_access_data->p_files ) free( p_access_data->p_files );

	if( p_access_data->p_gaps )
		free( p_access_data->p_gaps );
	if( p_access_data->p_avails )
		free( p_access_data->p_avails );
    free( p_access_data );
}

/*****************************************************************************
 * Read: standard read on a file descriptor.
 *****************************************************************************/
static ssize_t Read( input_thread_t * p_input, byte_t * p_buffer, size_t i_len )
{
    _input_socket_t * p_access_data = (_input_socket_t *)p_input->p_access_data;
    ssize_t i_ret;

	if( p_access_data->p_gaps )
	{
		if( p_access_data->l_fpos >= p_access_data->l_part_file_size )
		{
			DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), VLC_MSG_INFO, MODULE_STRING, "EOF at %u", p_access_data->l_fpos) );
			i_ret = 0;
		}
		else
		{
			size_t i_read = 0;

			i_ret = 0;
			while(   p_access_data->i_next_gap < p_access_data->i_gaps
				  && (   p_access_data->l_fpos                    >= p_access_data->p_gaps[p_access_data->i_next_gap].l_start
					  || p_access_data->l_fpos + (i_len - i_read) >  p_access_data->p_gaps[p_access_data->i_next_gap].l_start 
					 )
				  && i_read < i_len
			     )
			{
				ssize_t i_before_gap = p_access_data->p_gaps[p_access_data->i_next_gap].l_start - p_access_data->l_fpos;
				if( i_before_gap < 0 || i_before_gap > (ssize_t)(i_len - i_read) )
				{
					DEBUG_ONLY( MessageBeep(MB_ICONERROR) );
					msg_Err( p_input, "gap error at %u, i_before_gap=%d", p_access_data->l_fpos, i_before_gap);
					i_ret = -1;
					errno = EINVAL;
					break;
				}

				i_ret = read(p_access_data->_socket.i_handle, p_buffer, i_before_gap);
				DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), (i_ret>=0) ? VLC_MSG_INFO : VLC_MSG_ERR, MODULE_STRING, "reading %u bytes at %u", i_before_gap, p_access_data->l_fpos) );
				if( i_ret >= 0 )
				{
					if( i_ret != i_before_gap )
					{
						DEBUG_ONLY( MessageBeep(MB_ICONERROR) );
						DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), VLC_MSG_ERR, MODULE_STRING, "premature EOF; read %u bytes", i_ret) );
					}

					i_read += i_ret;
					p_buffer += i_ret;

					p_access_data->l_fpos = p_access_data->p_gaps[p_access_data->i_next_gap].l_end;
					DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), VLC_MSG_INFO, MODULE_STRING, "positioning to %u", p_access_data->l_fpos) );
					lseek(p_access_data->_socket.i_handle, p_access_data->l_fpos, SEEK_SET);
					p_access_data->i_next_gap++;
				}
				else
					break;
			}

			if( i_ret >= 0 && p_access_data->l_fpos < p_access_data->l_part_file_size )
			{
				ssize_t i_left = i_len - i_read;

				// don't let VLC read over the EOF
				if( i_left > 0 )
				{
					ssize_t i_file_left = p_access_data->l_part_file_size - p_access_data->l_fpos;
					if( i_file_left < i_left )
						i_left = i_file_left;
				}
				if( i_left > 0 )
				{
					i_ret = read(p_access_data->_socket.i_handle, p_buffer, i_left);
					DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), (i_ret>=0) ? VLC_MSG_INFO : VLC_MSG_ERR, MODULE_STRING, "reading %u bytes at %u", i_left, p_access_data->l_fpos) );
					if( i_ret > 0 )
					{
						if( i_ret != i_left )
						{
							DEBUG_ONLY( MessageBeep(MB_ICONERROR) );
							DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), VLC_MSG_ERR, MODULE_STRING, "premature EOF; read %u bytes", i_ret) );
						}

						i_read += i_ret;
						p_access_data->l_fpos += i_ret;
					}
				}
			}

			if( i_ret >= 0 )
				i_ret = i_read;
		}
	}
	else
	{
        /* b_pace_control || WIN32 */
		i_ret = read(p_access_data->_socket.i_handle, p_buffer, i_len);
	}

    if( i_ret < 0 )
    {
#ifdef HAVE_ERRNO_H
        if ( errno != EINTR && errno != EAGAIN )
            msg_Err( p_input, "read failed (%s)", strerror(errno) );
#else
        msg_Err( p_input, "read failed" );
#endif

		// workaround for a kinda strange bug in VLC (this bug does not only happen for part files but
		// also for "normal" files).. sometimes VLC creates and endless loop reading a file although the 'Read'
		// function returns an error all the time.. this locks up the entire system!!
		p_access_data->i_read_errors++;
		if( p_access_data->i_read_errors >= 3 )
		{
			msg_Err( p_input, "file reader forced EOF" );
			i_ret = 0; // indicate EOF to let VLC stop reading
		}
        /* Delay a bit to avoid consuming all the CPU. This is particularly
         * useful when reading from an unconnected FIFO. */
        msleep( INPUT_ERROR_SLEEP );
    }
	else
		p_access_data->i_read_errors = 0;

    p_access_data->i_nb_reads++;
#ifdef HAVE_SYS_STAT_H
    if ( p_input->stream.p_selected_area->i_size != 0
		    && (p_access_data->i_nb_reads % INPUT_FSTAT_NB_READS) == 0 
			&& !p_access_data->p_gaps )
    {
        struct stat stat_info;
        int i_file = p_access_data->i_index;

        if ( fstat( p_access_data->_socket.i_handle, &stat_info ) == -1 )
        {
#   ifdef HAVE_ERRNO_H
            msg_Warn( p_input, "couldn't stat again the file (%s)",
                      strerror(errno) );
#   else
            msg_Warn( p_input, "couldn't stat again the file" );
#   endif
        }
        else if ( p_access_data->p_files[i_file]->i_size != stat_info.st_size )
        {
            p_input->stream.p_selected_area->i_size +=
                (stat_info.st_size - p_access_data->p_files[i_file]->i_size);
            p_input->stream.b_changed = 1;
        }
    }
#endif

    /* If we reached an EOF then switch to the next file in the list */
    if ( i_ret == 0 && p_access_data->i_index + 1 < p_access_data->i_files )
    {
        int i_handle = p_access_data->_socket.i_handle;
        char *psz_name =
            p_access_data->p_files[++p_access_data->i_index]->psz_name;

        msg_Dbg( p_input, "opening file `%s'", psz_name );

        if ( _OpenFile( p_input, psz_name ) != VLC_SUCCESS )
        {
            p_access_data->_socket.i_handle = i_handle;
            return 0;
        }

        close( i_handle );

        /* We have to read some data */
        return Read( p_input, p_buffer, i_len );
    }

    return i_ret;
}

/*****************************************************************************
 * Seek: seek to a specific location in a file
 *****************************************************************************/
static void Seek( input_thread_t * p_input, off_t i_pos )
{
#define S p_input->stream
    _input_socket_t *p_access_data = (_input_socket_t *)p_input->p_access_data;
    int64_t i_size = 0;

    /* Check which file we need to access */
    if ( p_access_data->i_files > 1 )
    {
        int i, i_handle = p_access_data->_socket.i_handle;
        char * psz_name;

        for ( i = 0; i < p_access_data->i_files - 1; i++ )
        {
            if( i_pos < p_access_data->p_files[i]->i_size + i_size ) break;
            i_size += p_access_data->p_files[i]->i_size;
        }

        psz_name = p_access_data->p_files[i]->psz_name;

        msg_Dbg( p_input, "opening file `%s'", psz_name );
        if ( i != p_access_data->i_index &&
             _OpenFile( p_input, psz_name ) == VLC_SUCCESS )
        {
            /* Close old file */
            close( i_handle );
            p_access_data->i_index = i;
        }
        else
        {
            p_access_data->_socket.i_handle = i_handle;
        }
    }

	if( !p_access_data->p_gaps )
	{
#if defined( WIN32 ) && !defined( UNDER_CE )
		_lseeki64( p_access_data->_socket.i_handle, i_pos - i_size, SEEK_SET );
#else
		lseek( p_access_data->_socket.i_handle, i_pos - i_size, SEEK_SET );
#endif
	}
	else
	{
		int i;
		vlc_bool_t b_found = VLC_FALSE;

		// it depends on the file format how much 'seek' requests we'll get.. 
		// for *seekable* AVI files we nearly constantly have to seek.. 
		// actually for *seekable* AVI files each 64K block is read two(!) times..
		//DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), VLC_MSG_INFO, MODULE_STRING, "seek to stream pos %I64u", i_pos) );

		for( i = 0; i < p_access_data->i_avails; i++ )
		{
			if( (size_t)i_pos >= p_access_data->p_avails[i].l_stream_start && (size_t)i_pos < p_access_data->p_avails[i].l_stream_end )
			{
				b_found = VLC_TRUE;

				p_access_data->l_fpos = p_access_data->p_avails[i].l_start + ((size_t)i_pos - p_access_data->p_avails[i].l_stream_start);
				DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), VLC_MSG_INFO, MODULE_STRING, "seek to file pos %u (stream pos %u)", p_access_data->l_fpos, (size_t)i_pos) );
				lseek( p_access_data->_socket.i_handle, p_access_data->l_fpos, SEEK_SET );

				p_access_data->i_next_gap = p_access_data->p_avails[i].i_next_gap;

#ifdef _DEBUG
				{
					int i_next_gap = p_access_data->i_gaps;
					for( i = 0; i < p_access_data->i_gaps; i++ )
					{
						if( p_access_data->l_fpos <= p_access_data->p_gaps[i].l_start )
						{
							i_next_gap = i;
							break;
						}
					}

					if( i_next_gap != p_access_data->i_next_gap )
					{
						DEBUG_ONLY( MessageBeep(MB_ICONERROR) );
						DEBUG_ONLY( __msg_Generic( VLC_OBJECT(p_input), VLC_MSG_ERR, MODULE_STRING, "i_next_gap=%d  p_access_data->i_next_gap=%d", i_next_gap, p_access_data->i_next_gap) );
					}
				}
#endif
				break;
			}
		}

		if( !b_found )
		{
			DEBUG_ONLY( MessageBeep(MB_ICONERROR) );
	        msg_Err( p_input, "seeking at %u failed", (size_t)i_pos );
			return;
		}
	}

    vlc_mutex_lock( &S.stream_lock );
    S.p_selected_area->i_tell = i_pos;
    if( S.p_selected_area->i_tell > S.p_selected_area->i_size )
    {
        msg_Err( p_input, "seeking too far" );
        S.p_selected_area->i_tell = S.p_selected_area->i_size;
    }
    else if( S.p_selected_area->i_tell < 0 )
    {
        msg_Err( p_input, "seeking too early" );
        S.p_selected_area->i_tell = 0;
    }
    vlc_mutex_unlock( &S.stream_lock );
#undef S
}

/*****************************************************************************
 * OpenFile: Opens a specific file
 *****************************************************************************/
static int _OpenFile( input_thread_t * p_input, char * psz_name )
{
    input_socket_t * p_access_data = (input_socket_t *)p_input->p_access_data;

#ifdef UNDER_CE
    wchar_t psz_filename[MAX_PATH];
    MultiByteToWideChar( CP_ACP, 0, psz_name, -1, psz_filename, MAX_PATH );

    p_access_data->i_handle =
       (int)CreateFile( psz_filename, GENERIC_READ, FILE_SHARE_READ,
                        NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

    if ( (HANDLE)p_access_data->i_handle == INVALID_HANDLE_VALUE )
    {
        msg_Err( p_input, "cannot open file %s", psz_name );
        return VLC_EGENERIC;
    }
    p_input->stream.p_selected_area->i_size =
        GetFileSize( (HANDLE)p_access_data->i_handle, NULL );
#else

    p_access_data->i_handle = open( psz_name, O_BINARY | O_NONBLOCK /*| O_LARGEFILE*/ );
    if ( p_access_data->i_handle == -1 )
    {
#   ifdef HAVE_ERRNO_H
        msg_Err( p_input, "cannot open file %s (%s)", psz_name,
                 strerror(errno) );
#   else
        msg_Err( p_input, "cannot open file %s", psz_name );
#   endif
        return VLC_EGENERIC;
    }
#endif

    return VLC_SUCCESS;
}
